/*
 * Filesystem resource registry incorporating opening functionality.
 *
 * Copyright (C) 2021, 2022 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include "directory_resource.h"
#include "file_pager.h"
#include "resource_registry.h"

#include <fsclient/file.h>      /* file_region_flags */
#include <systypes/fcntl.h>



/* The file object registry provides a memory page collection to the page
   mappers it creates. */

ResourceRegistry::ResourceRegistry(Pages *pages)
: _pages(pages)
{
}



/* Convert opening flags to map-compatible paging flags. */

map_flags_t ResourceRegistry::get_flags(flags_t flags)
{
    return file_region_flags(flags);
}



/* Make a directory provider. */

long ResourceRegistry::make_directory_provider(FileOpening *opening,
                                               flags_t flags,
                                               fileid_t fileid,
                                               Provider **provider)
{
    /* Make an accessor and a provider to encapsulate it. */

    DirectoryAccessor *accessor;
    long err = opening->make_directory_accessor(flags, fileid, &accessor);

    if (err)
        return err;

    *provider = new DirectoryProvider(fileid, this, accessor);
    return L4_EOK;
}

/* Make a file provider. */

long ResourceRegistry::make_file_provider(FileOpening *opening,
                                          flags_t flags,
                                          fileid_t fileid,
                                          Provider **provider)
{
    /* Make an accessor, page mapper, and a provider to encapsulate them. */

    Accessor *accessor;
    long err = opening->make_accessor(flags, fileid, &accessor);

    if (err)
        return err;

    PageMapper *mapper = new PageMapper(accessor, _pages);
    *provider = new FileProvider(fileid, this, mapper);
    return L4_EOK;
}

/* Make a provider of the appropriate type. */

long ResourceRegistry::make_provider(FileOpening *opening,
                                     flags_t flags,
                                     fileid_t fileid,
                                     Provider **provider)
{
    long err = -L4_EIO;

    if (opening->accessing_directory(flags, fileid))
        err = make_directory_provider(opening, flags, fileid, provider);
    else if (opening->accessing_file(flags, fileid))
        err = make_file_provider(opening, flags, fileid, provider);

    if (err)
        return err;

    /* Register the provider. */

    set(fileid, *provider);
    return L4_EOK;
}



/* Obtain any active provider for the given 'fileid'. */

long ResourceRegistry::find_provider(fileid_t fileid, Provider **provider)
{
    /* Obtain any registered provider. */

    Accountable *accountable = get(fileid);

    if (accountable != NULL)
    {
        *provider = dynamic_cast<Provider *>(accountable);

        if ((*provider) != NULL)
            return L4_EOK;
        else
            return -L4_EIO;
    }

    return -L4_ENOENT;
}



/* Return a resource for a filesystem object. */

long ResourceRegistry::get_resource(FileOpening *opening, const char *path,
                                    flags_t flags, offset_t *size,
                                    object_flags_t *object_flags,
                                    Resource **resource)
{
    std::lock_guard<std::mutex> guard(_lock);

    /* Obtain an identifier for the object, even for new files (subject to use
       of the appropriate flag). This also checks the access rights for any
       configured user. */

    fileid_t fileid;
    long err = opening->get_fileid(path, flags, &fileid);

    if (err)
        return err;

    /* Obtain a provider for the object. */

    Provider *provider;
    err = find_provider(fileid, &provider);

    /* Make a new provider if necessary. */

    if (err == -L4_ENOENT)
        err = make_provider(opening, flags, fileid, &provider);

    if (err)
        return err;

    /* Make a resource for the provider. */

    return provider->make_resource(get_flags(flags), size, object_flags, resource);
}

/* Request the removal of a filesystem object through any active provider or
   perform the direct removal of any inactive object. */

long ResourceRegistry::remove_provider(FileOpening *opening, const char *path)
{
    std::lock_guard<std::mutex> guard(_lock);

    /* Obtain an identifier for any recognised object. This also checks the
       access rights for any configured user. */

    fileid_t fileid;
    long err = opening->get_fileid(path, 0, &fileid);

    if (err)
        return err;

    return _remove_provider(opening, path, fileid);
}

long ResourceRegistry::_remove_provider(FileOpening *opening, const char *path,
                                        fileid_t fileid)
{
    /* Obtain a provider for the object. */

    Provider *provider;
    long err = find_provider(fileid, &provider), lookup_err = err;

    if (err && (err != -L4_ENOENT))
        return err;

    fileid_t parent_fileid;

    err = opening->get_parent(path, &parent_fileid);

    if (err)
        return err;

    /* Test for an empty directory.
       NOTE: A descriptive error should be communicated. */

    if (opening->accessing_directory(0, fileid) && !opening->directory_is_empty(fileid))
        return -L4_EIO;

    /* Unlink the object regardless of whether it will be removed. */

    err = opening->unlink_object(parent_fileid, fileid);

    if (err)
        return err;

    /* Without a provider being active, remove the object directly. */

    if (lookup_err)
        return opening->remove_object(fileid);

    /* With a provider active, request the object's removal upon closure. */

    provider->remove_pending(true);
    return L4_EOK;
}

/* Request the renaming of a filesystem object. */

long ResourceRegistry::rename_provider(FileOpening *opening, const char *source,
                                       const char *target)
{
    std::lock_guard<std::mutex> guard(_lock);

    long err;
    fileid_t source_fileid, target_fileid;

    /* Obtain an identifier for any recognised object. This also checks the
       access rights for any configured user. */

    err = opening->get_fileid(source, 0, &source_fileid);

    if (err)
        return err;

    /* If source is a directory, the target must be a new directory or an empty
       directory. An empty directory will be replaced by the source
       directory. */

    if (opening->accessing_directory(0, source_fileid))
    {
        err = opening->get_fileid(target, 0, &target_fileid);

        if (err && (err != -L4_ENOENT))
            return err;

        /* Any existing directory must be empty. */

        if (!err)
        {
            /* NOTE: Should really use the equivalent of ENOTDIR. */

            if (!opening->accessing_directory(0, target_fileid))
                return -L4_EIO;

            /* NOTE: Test for empty directory. */

            /* The existing directory will be removed. */

            err = _remove_provider(opening, target, target_fileid);

            if (err)
                return err;
        }

        /* The source directory is moved within the filesystem to the parent of
           the indicated target. */

        return opening->rename_object(source, target);
    }

    /* If source is a file, the target must be a new or existing file, not a
       directory. An existing file will be replaced. */

    else if (opening->accessing_file(0, source_fileid))
    {
        err = opening->get_fileid(target, 0, &target_fileid);

        if (err && (err != -L4_ENOENT))
            return err;

        if (!err)
        {
            /* NOTE: Should really use the equivalent of EISDIR. */

            if (!opening->accessing_file(0, target_fileid))
                return -L4_EIO;

            /* The existing file will be removed. */

            err = _remove_provider(opening, target, target_fileid);

            if (err)
                return err;
        }

        /* The source file is moved within the filesystem to the parent of the
           indicated target. */

        return opening->rename_object(source, target);
    }

    /* NOTE: Other object types are to be supported. */

    else
        return -L4_EIO;
}

// vim: tabstop=4 expandtab shiftwidth=4
